/*
 * Copyright (C) 2013 Texas Instruments Incorporated - http://www.ti.com/
 *
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <errno.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#include "cJSON/cJSON.h"
#include "mpm_transport_cfg.h"
#include "mpm_transport_sharedmem.h"
#include "mpm_transport_hyplnk.h"

#define DEFAULT_CONFIG_FILE_NAME "/etc/mpm/mpm_config.json"

#define json_obj_getnum(obj, item, iteme) \
	do { \
		if (obj->type == cJSON_Number) { \
			item->iteme = obj->valueint; \
		} else if (obj->type == cJSON_String) { \
			item->iteme = strtoull(obj->valuestring, NULL, 0); \
		} else { \
			mpm_printf(1, "Invalid type (%d) for segment %s\n", \
					obj->type, item->name); \
			return (-1); \
		} \
	} while(0)

int json_get_mmap_info (cJSON *json, char *map_name, mpm_transport_cfg_t *sp, int map_index)
{
	mpm_mmap_t *mmap = &(sp->mmap[map_index]);
	cJSON * segs;
	cJSON * segp;
	cJSON * sege;
	int len, i;

	segs = cJSON_GetObjectItem(json,"segments");
	if (!segs) {
		mpm_printf(1, "No segments object array in cfg file\n");
		return (-1);
	}

	if (segs->type != cJSON_Array) {
		mpm_printf(1, "Segments must be array type\n");
		return (-1);
	}

	len = cJSON_GetArraySize(segs);
	if (len > MPM_MAX_SEGMENTS) {
		mpm_printf(1, "The number of segments (%d) must be less than %d\n",
				len, MPM_MAX_SEGMENTS);
		return (-1);
	}
	for (i = 0; i < len; i++) {
		segp = cJSON_GetArrayItem(segs, i);
		if (!segp) {
			mpm_printf(1, "Could not read segment for index %d\n", i);
			return (-1);
		}
		sege = cJSON_GetObjectItem(segp, "name");
		if (!sege) {
			mpm_printf(1, "Segment without a name, skipping\n");
			continue;
		}

		if (strlen(sege->valuestring) > MPM_MAX_NAME_LENGTH) {
			mpm_printf(1, "Segment name is too big (%d), must be less than %d\n",
					(int) strlen(sege->valuestring), MPM_MAX_NAME_LENGTH);
			return (-1);
		}

		if (strncmp(sege->valuestring, map_name, MPM_MAX_NAME_LENGTH)) {
			continue;
		}

		strncpy(mmap->name, sege->valuestring, MPM_MAX_NAME_LENGTH);

		sege = cJSON_GetObjectItem(segp, "globaladdr");
		if (!sege) {
			mpm_printf(1, "No global address for segment %s\n", mmap->name);
			return (-1);
		}
		json_obj_getnum(sege, mmap, global_addr);

		sege = cJSON_GetObjectItem(segp, "length");
		if (!sege) {
			mpm_printf(1, "No length for segment %s\n", mmap->name);
			return (-1);
		}
		
		json_obj_getnum(sege, mmap, length);
		
		sege = cJSON_GetObjectItem(segp, "localaddr");
		if (!sege) {
			mmap->local_addr = mmap->global_addr;
		} else {
			json_obj_getnum(sege, mmap, local_addr);
		}

		if (sp->transport == shared_memory) {
			/* This section to be moved to sharedmem directory */
			sege = cJSON_GetObjectItem(segp, "devicename");
			if (!sege) {
				mpm_printf(1, "Need device name for shared memory transport\n");
				return (-1);
			}

			if (strlen(sege->valuestring) > MPM_MAX_NAME_LENGTH) {
				mpm_printf(1, "Segment device name is too big (%d), must be less than %d\n",
						(int) strlen(sege->valuestring), MPM_MAX_NAME_LENGTH);
				return (-1);
			}
			strncpy(mmap->devicename, sege->valuestring, MPM_MAX_NAME_LENGTH);
		}
		if (sp->transport == hyperlink) {
			/* This section to be moved to hyplnk directory */
			sege = cJSON_GetObjectItem(segp, "devicename");
			if (!sege) {
				mpm_printf(1, "Need device name for hyperlink transport\n");
				return (-1);
			}

			if (strlen(sege->valuestring) > MPM_MAX_NAME_LENGTH) {
				mpm_printf(1, "Segment device name is too big (%d), must be less than %d\n",
						(int) strlen(sege->valuestring), MPM_MAX_NAME_LENGTH);
				return (-1);
			}
			strncpy(mmap->devicename, sege->valuestring, MPM_MAX_NAME_LENGTH);
		}

		break;
	}

	return (0);
}

int json_get_dma_info (cJSON *json, char *dma_name, mpm_transport_cfg_t *sp)
{
	cJSON * segs;
	cJSON * segp;
	cJSON * sege;
	int len, i;

	segs = cJSON_GetObjectItem(json,"dma-params");
	if (!segs) {
		mpm_printf(1, "No dma-params object array in cfg file\n");
		return (-1);
	}

	if (segs->type != cJSON_Array) {
		mpm_printf(1, "dma-params must be array type\n");
		return (-1);
	}

	len = cJSON_GetArraySize(segs);
	if (len > MPM_MAX_DMA_PARAMS) {
		mpm_printf(1, "The number of dma-params (%d) must be less than %d\n",
				len, MPM_MAX_DMA_PARAMS);
		return (-1);
	}
	for (i = 0; i < len; i++) {
		segp = cJSON_GetArrayItem(segs, i);
		if (!segp) {
			mpm_printf(1, "Could not read segment for index %d\n", i);
			return (-1);
		}
		sege = cJSON_GetObjectItem(segp, "name");
		if (!sege) {
			mpm_printf(1, "DMA profile without a name, skipping\n");
			continue;
		}

		if (strlen(sege->valuestring) > MPM_MAX_NAME_LENGTH) {
			mpm_printf(1, "DMA profile name is too big (%d), must be less than %d\n",
					(int) strlen(sege->valuestring), MPM_MAX_NAME_LENGTH);
			return (-1);
		}

		if (strncmp(sege->valuestring, dma_name, MPM_MAX_NAME_LENGTH)) {
			continue;
		}

		sege = cJSON_GetObjectItem(segp, "edma3inst");
		if (!sege) {
			mpm_printf(1, "No edma3inst specified for %s\n", sege->valuestring);
			return (-1);
		}
		else {
			json_obj_getnum(sege, sp, edma3_inst);
		}

		sege = cJSON_GetObjectItem(segp, "shadowregion");
		if (!sege) {
			mpm_printf(1, "No shadowregion specified for %s\n", sege->valuestring);
			return (-1);
		}
		else {
			json_obj_getnum(sege, sp, shadow_reg);
		}

		sege = cJSON_GetObjectItem(segp, "prealloc");
		if (!sege) {
			sp->edma3_prealloc = 0;
		}
		else {
			json_obj_getnum(sege, sp, edma3_prealloc);
		}
	}
	return 0;
}

int json_get_mpax_info (cJSON *json, char *mpax_name, mpm_transport_cfg_t *sp)
{
	cJSON * segs;
	cJSON * segp;
	cJSON * sege;
	int len, i;

	segs = cJSON_GetObjectItem(json,"mpax-params");
	if (!segs) {
		mpm_printf(1, "No mpax-params object array in cfg file\n");
		return (-1);
	}

	if (segs->type != cJSON_Array) {
		mpm_printf(1, "mpax-params must be array type\n");
		return (-1);
	}

	len = cJSON_GetArraySize(segs);
	if (len > MPM_MAX_MPAX_PARAMS) {
		mpm_printf(1, "The number of mpax-params (%d) must be less than %d\n",
				len, MPM_MAX_MPAX_PARAMS);
		return (-1);
	}
	for (i = 0; i < len; i++) {
		segp = cJSON_GetArrayItem(segs, i);
		if (!segp) {
			mpm_printf(1, "Could not read segment for index %d\n", i);
			return (-1);
		}
		sege = cJSON_GetObjectItem(segp, "name");
		if (!sege) {
			mpm_printf(1, "mpax profile without a name, skipping\n");
			continue;
		}

		if (strlen(sege->valuestring) > MPM_MAX_NAME_LENGTH) {
			mpm_printf(1, "mpax profile name is too big (%d), must be less than %d\n",
					(int) strlen(sege->valuestring), MPM_MAX_NAME_LENGTH);
			return (-1);
		}

		if (strncmp(sege->valuestring, mpax_name, MPM_MAX_NAME_LENGTH)) {
			continue;
		}

		sege = cJSON_GetObjectItem(segp, "base");
		if (!sege) {
			mpm_printf(1, "No base specified for %s\n", sege->valuestring);
			return (-1);
		}
		else {
			json_obj_getnum(sege, sp, mpax_base);
		}

		sege = cJSON_GetObjectItem(segp, "size");
		if (!sege) {
			mpm_printf(1, "No size specified for %s\n", sege->valuestring);
			return (-1);
		}
		else {
			json_obj_getnum(sege, sp, mpax_size);
		}

		sege = cJSON_GetObjectItem(segp, "index");
		if (!sege) {
			mpm_printf(1, "No index specified for %s\n", sege->valuestring);
			return (-1);
		}
		else {
			json_obj_getnum(sege, sp, mpax_index);
		}

		sege = cJSON_GetObjectItem(segp, "count");
		if (!sege) {
			mpm_printf(1, "No count specified for %s\n", sege->valuestring);
			return (-1);
		}
		else {
			json_obj_getnum(sege, sp, mpax_count);
		}
	}
	return 0;
}

int json_get_transport_info (cJSON *json, char *transport_name, mpm_transport_cfg_t *sp)
{
	mpm_transport_hyplnk_t * hypInfo;
	cJSON * segs;
	cJSON * segp;
	cJSON * sege;
	int len, i;

	segs = cJSON_GetObjectItem(json,"transports");
	if (!segs) {
		mpm_printf(1, "No transports object array in cfg file\n");
		return (-1);
	}

	if (segs->type != cJSON_Array) {
		mpm_printf(1, "Transports must be array type\n");
		return (-1);
	}

	len = cJSON_GetArraySize(segs);
	if (len > MPM_MAX_TRANSPORTS) {
		mpm_printf(1, "The number of transports (%d) must be less than %d\n",
				len, MPM_MAX_TRANSPORTS);
		return (-1);
	}

	if (!strcmp(transport_name, "sharedmemory")) {
		sp->transport = shared_memory;
		return 0;
	}

	for (i = 0; i < len; i++) {
		segp = cJSON_GetArrayItem(segs, i);
		if (!segp) {
			mpm_printf(1, "Could not read segment for index %d\n", i);
			return (-1);
		}
		sege = cJSON_GetObjectItem(segp, "name");
		if (!sege) {
			mpm_printf(1, "Transport without a name, skipping\n");
			continue;
		}

		if (strlen(sege->valuestring) > MPM_MAX_NAME_LENGTH) {
			mpm_printf(1, "Transport name is too big (%d), must be less than %d\n",
					(int) strlen(sege->valuestring), MPM_MAX_NAME_LENGTH);
			return (-1);
		}

		if (strncmp(sege->valuestring, transport_name, MPM_MAX_NAME_LENGTH)) {
			continue;
		}

		sege = cJSON_GetObjectItem(segp, "transporttype");
		if (!sege) {
			mpm_printf(1, "No transporttype specified for %s\n", sege->valuestring);
			return (-1);
		}
		else {
			if (!strcmp(sege->valuestring, "sharedmemory")) {
				sp->transport = shared_memory;
			} else if (!strcmp(sege->valuestring, "pcie")) {
				sp->transport = pcie;
			} else if (!strcmp(sege->valuestring, "srio")) {
				sp->transport = srio;
			} else if (!strcmp(sege->valuestring, "ethernet")) {
				sp->transport = ethernet;
			} else if (!strcmp(sege->valuestring, "hyperlink")) {
				sp->transport = hyperlink;
			} else {
				mpm_printf(1, "Invalid transport %s in configuration\n",
						sege->valuestring);
				return (-1);
			}
		}

		switch (sp->transport) {
			case shared_memory:
				break;
			case hyperlink:
				hypInfo = calloc(1, sizeof(mpm_transport_hyplnk_t));
				if (!hypInfo) {
					mpm_printf(1, "can't allocate memory for hyperlink data \n");
					return -1;
				}
				else
				{
					strcpy(hypInfo->hyplnkSettings, transport_name);
					strcpy(hypInfo->name, transport_name);
					sege = cJSON_GetObjectItem(segp, "direction");
					if (!sege)
					{
						mpm_printf(1, "No hyperlink direction given\n");
						return (-1);
					}
					else
					{
						if (!strcmp(sege->valuestring, "loopback")) {
							hypInfo->hyplnkDir = hyplnk_LOOPBACK;
						} else if (!strcmp(sege->valuestring, "remote")) {
							hypInfo->hyplnkDir = hyplnk_REMOTE;}
						else
						{
							mpm_printf(1, "hyperlink direction not supported\n");
							return (-1);
						}
					}

					sege = cJSON_GetObjectItem(segp, "hyplnkinterface");
					if (!sege)
					{
						mpm_printf(1, "No hyperlink interface given\n");
						return (-1);
					}
					else
					{
						if (!strcmp(sege->valuestring, "hyplnk0")) {
							hypInfo->hyplnkIF = hyplnk_PORT_0;
						} else if (!strcmp(sege->valuestring, "hyplnk1")) {
							hypInfo->hyplnkIF = hyplnk_PORT_1;}
						else
						{
							mpm_printf(1, "hyperlink interface not supported\n");
							return (-1);
						}
					}

					sege = cJSON_GetObjectItem(segp, "txprivid");
					if (!sege)
					{
						mpm_printf(1, "No hyperlink txprivid size given\n");
						return (-1);
					}
					else
					{
						json_obj_getnum(sege, hypInfo, hyplnkTxPrivID);
					}

					sege = cJSON_GetObjectItem(segp, "rxprivid");
					if (!sege)
					{
						mpm_printf(1, "No hyperlink rxprivid size given\n");
						return (-1);
					}
					else
					{
						json_obj_getnum(sege, hypInfo, hyplnkRxPrivID);
					}

					sege = cJSON_GetObjectItem(segp, "rxsegsel");
					if (!sege)
					{
						mpm_printf(1, "No hyperlink rxsegsel size given\n");
						return (-1);
					}
					else
					{
						json_obj_getnum(sege, hypInfo, hyplnkSegSel);
					}

					sege = cJSON_GetObjectItem(segp, "rxlenval");
					if (!sege)
					{
						mpm_printf(1, "No hyperlink rxlenval size given\n");
						return (-1);
					}
					else
					{
						json_obj_getnum(sege, hypInfo, hyplnkRxLen);
					}

					sege = cJSON_GetObjectItem(segp, "numlanes");
					if (!sege)
					{
						mpm_printf(1, "No hyperlink number of lanes given\n");
						return (-1);
					}
					else
					{
						json_obj_getnum(sege, hypInfo, hyplnkNumLanes);
					}
					if ((hypInfo->hyplnkNumLanes != 4) && (hypInfo->hyplnkNumLanes != 1))
					{
						mpm_printf(1, "Only 4 and 1 lane supported\n");
						return (-1);
					}
				}
				sp->td = (void *) hypInfo;
				break;
			default:
				break;
		}

		break;
	}

	return (0);
}

int json_get_slave_info (cJSON *json, char *slave_name, mpm_transport_cfg_t *sp)
{
	char *namep;
	int len, i;
        int lenm, m;
	cJSON *slvs, *slvp, *slve, *slvm;

	slvs = cJSON_GetObjectItem(json,"slaves");
	if (!slvs) {
		mpm_printf(1, "No slaves object array in cfg file\n");
		return (-1);
	}

	if (slvs->type != cJSON_Array) {
		mpm_printf(1, "Slaves must be array type\n");
		return (-1);
	}

	len = cJSON_GetArraySize(slvs);
	for (i = 0; i < len; i++) {
		slvp = cJSON_GetArrayItem(slvs, i);
		if (!slvp) {
			mpm_printf(1, "Could not read slave info for index %d\n", i);
			return (-1);
		}

		slve = cJSON_GetObjectItem(slvp, "name");
		if (!slve) {
			mpm_printf(1, "Slave processor without a name, skipping\n");
			continue;
		}

		if (strlen(slve->valuestring) > MPM_MAX_NAME_LENGTH) {
			mpm_printf(1, "Slave core name is too big (%d), must be less than %d\n",
					(int) strlen(slve->valuestring), MPM_MAX_NAME_LENGTH);
			continue;
		}

		if (strncmp(slve->valuestring, slave_name, MPM_MAX_NAME_LENGTH)) {
			continue;
		}

		strncpy(sp->name, slve->valuestring, MPM_MAX_NAME_LENGTH);

		slve = cJSON_GetObjectItem(slvp, "coreid");
		if (slve) {
			json_obj_getnum(slve, sp, core_id);
		}

		slve = cJSON_GetObjectItem(slvp, "clusterid");
		if (slve) {
			json_obj_getnum(slve, sp, cluster_id);
		}

		slve = cJSON_GetObjectItem(slvp, "transport");
		if (!slve) {
			mpm_printf(1, "No transport provided for slave %s\n",sp->name);
			return (-1);
		}
		if (json_get_transport_info(json, slve->valuestring, sp)) {
			mpm_printf(1, "Could not read transport info for slave %s\n", sp->name);
			return (-1);
		}

		slve = cJSON_GetObjectItem(slvp, "dma");
		if (slve) {
			strncpy(sp->edma3_name, slve->valuestring, MPM_MAX_NAME_LENGTH);
			if (json_get_dma_info(json, slve->valuestring, sp)) {
				mpm_printf(1, "Could not read dma info for slave %s\n", sp->name);
				return (-1);
			}
		}

		slve = cJSON_GetObjectItem(slvp, "mpax");
		if (slve) {
			strncpy(sp->mpax_name, slve->valuestring, MPM_MAX_NAME_LENGTH);
			if (json_get_mpax_info(json, slve->valuestring, sp)) {
				mpm_printf(1, "Could not read mpax info for slave %s\n", sp->name);
				return (-1);
			}
		}

		slvm = cJSON_GetObjectItem(slvp, "memorymap");

		if (!slvm) {
			mpm_printf(1, "No slave memory map in cfg file\n");
			return (-1);
		}

		if (slvm->type != cJSON_Array) {
			mpm_printf(1, "Slave memory map must be array type\n");
			return (-1);
		}

		lenm = cJSON_GetArraySize(slvm);
		if (lenm > MPM_MAX_MMAPS) {
			mpm_printf(1, "The number of memory map sections (%d) must be less than %d\n",
					lenm, MPM_MAX_MMAPS);
			return (-1);
		}

		for (m = 0; m < lenm; m++) {
			slvp = cJSON_GetArrayItem(slvm, m);
			if (!slvp) {
				mpm_printf(1, "Could not read map name for the index %d\n", m);
				return (-1);
			}

			namep = slvp->valuestring;
			if (json_get_mmap_info(json, namep, sp, m)) {
				mpm_printf(1, "Could not read slave info for index %d\n", m);
				return (-1);
			}
		}
		sp->num_mmap = lenm;

		break;
	}

	return (0);
}

/*
	slave_name:  should match the name in config file.
	cfg: runtime configuration needed for this device.
	In case of shared memory it might not have anything.
	Returns mpm_transport_h or NULL on error.
*/
mpm_transport_h mpm_transport_open(char *slave_name, mpm_transport_open_t *cfg)
{
	mpm_transport_cfg_t *sp = NULL;
	char *config_buf = NULL;
	cJSON *json = NULL;
	int retval = -1;
	FILE *fpr = NULL;
	struct stat buf;
	int flen;
	int res;
	
	if (!slave_name || !cfg) {
		mpm_printf(1, "Null parameter\n");
		goto close_n_exit;
	}
	if (strlen(slave_name) > MPM_MAX_NAME_LENGTH) {
		mpm_printf(1, "Name should be less than %d\n", MPM_MAX_NAME_LENGTH);
		goto close_n_exit;
	}

	sp = calloc(1, sizeof(mpm_transport_cfg_t));
	if (!sp) {
		mpm_printf(1, "can't allocate memory for handle\n");
		goto close_n_exit;
	}

	fpr = fopen(DEFAULT_CONFIG_FILE_NAME, "r");
	if (!fpr){
		mpm_printf(1, "Error opening file %s (error: %s)\n",
				DEFAULT_CONFIG_FILE_NAME, strerror(errno));
		goto close_n_exit;
	}

	fstat(fileno(fpr), &buf);
	flen = buf.st_size;

	config_buf = (char *) malloc(flen + 1);
	if (!config_buf) {
		mpm_printf(1, "Error in allocating cfg buffer (error: %s)\n",
				strerror(errno));
		goto close_n_exit;
	}
	config_buf[flen] = 0;

	res = fread (config_buf, 1, flen, fpr);
	if (res != flen) {
		mpm_printf(1, "Error in reading cfg file (error: %s)\n",
				strerror(errno));
		goto close_n_exit;
	}

	json = cJSON_Parse(config_buf);
	if (!json) {
		mpm_printf(1, "JSON parser error before: [%s]\n",cJSON_GetErrorPtr());
		goto close_n_exit;
	}

	if (json_get_slave_info(json, slave_name, sp)) {
		mpm_printf(1, "Error in JSON parsing\n");
		goto close_n_exit;
	}

	switch (sp->transport) {
		case shared_memory:
			if (mpm_transport_sharedmem_open(sp, cfg)) {
				mpm_printf(1, "shm open failed\n");
				goto close_n_exit;
			}
			break;
		case hyperlink:
			if (mpm_transport_hyplnk_open(sp, cfg)) {
				mpm_printf(1, "hyplnk open failed\n");
				goto close_n_exit;
			}
			break;
		default:
			mpm_printf(1, "unsupported transport %d\n", sp->transport);
			goto close_n_exit;
	}

	retval = 0;

close_n_exit:
	if(json) cJSON_Delete(json);
	if (config_buf) free(config_buf);
	if (fpr) fclose(fpr);

	if (retval < 0)
		free(sp);

	return ((retval < 0) ?  (mpm_transport_h) 0 : sp);
}

/*
	Read from slave[DSP] local address and length into the buffer.
	Returns 0 on success, error code on error.
*/
int mpm_transport_read(mpm_transport_h h, uint32_t addr, uint32_t length, char *buf, mpm_transport_read_t *cfg)
{
	mpm_transport_cfg_t *sp = (mpm_transport_cfg_t *)h;
	if (!h) {
		mpm_printf(1, "invalid handle \n");
		return -1;
	}
	switch (sp->transport) {
		case shared_memory:
			if (mpm_transport_sharedmem_read(sp, addr, length, buf, cfg)) {
				mpm_printf(1, "shm read failed\n");
				return -1;
			}
			break;
		case hyperlink:
			if (mpm_transport_hyplnk_read(sp, addr, length, buf, cfg)) {
				mpm_printf(1, "hyplnk read failed\n");
				return -1;
			}
			break;
		default:
			mpm_printf(1, "unsupported transport %d\n", sp->transport);
			return -1;
	}
	return 0;
}

/*
	Similar to mpm_transport_read, except memcpy is replaced with EDMA3 operation.
	Returns 0 on success, error code on error.
*/
mpm_transport_trans_h mpm_transport_get_initiate(mpm_transport_h h, uint32_t from_addr, uint32_t to_addr, uint32_t length, bool is_blocking, mpm_transport_read_t *cfg)
{
	mpm_transport_cfg_t *sp = (mpm_transport_cfg_t *)h;
	mpm_transport_trans_t *tp = (mpm_transport_trans_t *) NULL;
	if (!h) {
		mpm_printf(1, "invalid handle \n");
		return tp;
	}
	switch (sp->transport) {
		case shared_memory:
			mpm_printf(1, "shared_mem dma functions not supported\n");
			break;
		case hyperlink:
			tp = mpm_transport_hyplnk_get_initiate(sp, from_addr, to_addr, length, is_blocking, cfg);
			if (tp->key == -1)
				mpm_printf(1, "hyplnk dma read failed\n");
			else
				mpm_printf(2, "got key: 0x%08x\n",tp->key);
			break;
		default:
			mpm_printf(1, "unsupported transport %d\n", sp->transport);
	}
	return tp;
}

/*
	Write to slave[DSP] local address "addr" with "length" from the buffer "buf".
	Returns 0 on success, error code on error.
*/
int mpm_transport_write(mpm_transport_h h, uint32_t addr, uint32_t length, char *buf, mpm_transport_write_t *cfg)
{
	mpm_transport_cfg_t *sp = (mpm_transport_cfg_t *)h;
	if (!h) {
		mpm_printf(1, "invalid handle \n");
		return -1;
	}
	switch (sp->transport) {
		case shared_memory:
			if (mpm_transport_sharedmem_write(sp, addr, length, buf, cfg)) {
				mpm_printf(1, "shm write failed\n");
				return -1;
			}
			break;
		case hyperlink:
			if (mpm_transport_hyplnk_write(sp, addr, length, buf, cfg)) {
				mpm_printf(1, "hyplnk write failed\n");
				return -1;
			}
			break;
		default:
			mpm_printf(1, "unsupported transport %d\n", sp->transport);
			return -1;
	}
	return 0;
}

/*
	Similar to mpm_transport_write, except memcpy is replaced with EDMA3 operation.
	Returns 0 on success, error code on error.
*/
mpm_transport_trans_h mpm_transport_put_initiate(mpm_transport_h h, uint32_t to_addr, uint32_t from_addr, uint32_t length, bool is_blocking, mpm_transport_write_t *cfg)
{
	mpm_transport_cfg_t *sp = (mpm_transport_cfg_t *)h;
	mpm_transport_trans_t *tp = (mpm_transport_trans_t *) NULL;
	if (!h) {
		mpm_printf(1, "invalid handle \n");
		return tp;
	}
	switch (sp->transport) {
		case shared_memory:
			mpm_printf(1, "shared_mem dma functions not supported\n");
			break;
		case hyperlink:
			tp = mpm_transport_hyplnk_put_initiate(sp, to_addr, from_addr, length, is_blocking, cfg);
			if (tp->key == -1) {
				mpm_printf(1, "hyplnk dma write failed\n");
				tp = NULL;
			}
			else
				mpm_printf(2, "got key: 0x%08x\n",tp->key);
			break;
		default:
			mpm_printf(1, "unsupported transport %d\n", sp->transport);
	}
	return tp;
}

/*
	Similar to mpm_transport_get_initiate(), but with multiple transfers linked over EDMA3.
	Returns 0 on success, error code on error.
*/
mpm_transport_trans_h mpm_transport_get_initiate_linked(mpm_transport_h h, uint32_t num_links, uint32_t from_addr[], uint32_t to_addr[], uint32_t length[], bool is_blocking, mpm_transport_write_t *cfg)
{
	mpm_transport_cfg_t *sp = (mpm_transport_cfg_t *)h;
	mpm_transport_trans_t *tp = (mpm_transport_trans_t *) NULL;
	if (!h) {
		mpm_printf(1, "invalid handle \n");
		return tp;
	}
	switch (sp->transport) {
		case shared_memory:
			mpm_printf(1, "shared_mem dma functions not supported\n");
			break;
		case hyperlink:
			tp = mpm_transport_hyplnk_get_initiate_linked(sp, num_links, from_addr, to_addr, length, is_blocking, cfg);
			if (tp->key == -1) {
				mpm_printf(1, "hyplnk dma write failed\n");
				tp = NULL;
			}
			else
				mpm_printf(2, "got key: 0x%08x\n",tp->key);
			break;
		default:
			mpm_printf(1, "unsupported transport %d\n", sp->transport);
	}
	return tp;
}

/*
	Similar to mpm_transport_put_initiate(), but with multiple transfers linked over EDMA3.
	Returns 0 on success, error code on error.
*/
mpm_transport_trans_h mpm_transport_put_initiate_linked(mpm_transport_h h, uint32_t num_links, uint32_t to_addr[], uint32_t from_addr[], uint32_t length[], bool is_blocking, mpm_transport_write_t *cfg)
{
	mpm_transport_cfg_t *sp = (mpm_transport_cfg_t *)h;
	mpm_transport_trans_t *tp = (mpm_transport_trans_t *) NULL;
	if (!h) {
		mpm_printf(1, "invalid handle \n");
		return tp;
	}
	switch (sp->transport) {
		case shared_memory:
			mpm_printf(1, "shared_mem dma functions not supported\n");
			break;
		case hyperlink:
			tp = mpm_transport_hyplnk_put_initiate_linked(sp, num_links, to_addr, from_addr, length, is_blocking, cfg);
			if (tp->key == -1) {
				mpm_printf(1, "hyplnk dma write failed\n");
				tp = NULL;
			}
			else
				mpm_printf(2, "got key: 0x%08x\n",tp->key);
			break;
		default:
			mpm_printf(1, "unsupported transport %d\n", sp->transport);
	}
	return tp;
}

/*
	Checks for EDMA completion. Clears the pending interrupt if complete.
	Returns:
	   -1 = error
		0 = In progress
		1 = complete
*/
int mpm_transport_transfer_check(mpm_transport_h h, mpm_transport_trans_h th)
{
	mpm_transport_cfg_t *sp = (mpm_transport_cfg_t *)h;
	int ret = -1;
	if (!h) {
		mpm_printf(1, "invalid handle \n");
		return -1;
	}
	switch (sp->transport) {
		case shared_memory:
			mpm_printf(1, "shared_mem dma functions not supported\n");
			break;
		case hyperlink:
			ret = mpm_transport_hyplnk_transfer_check(sp, th);
			break;
		default:
			mpm_printf(1, "unsupported transport %d\n", sp->transport);
			return -1;
	}
	return ret;
}

/*
	Creates a mmap for the memory location and returns virtual address.
*/
void *mpm_transport_mmap(mpm_transport_h h, uint32_t addr, uint32_t length, mpm_transport_mmap_t *cfg)
{
	mpm_transport_cfg_t *sp = (mpm_transport_cfg_t *)h;
	if (!h) {
		mpm_printf(1, "invalid handle \n");
		return (void *) -1;
	}
	switch (sp->transport) {
		case shared_memory:
			return (mpm_transport_sharedmem_mmap(sp, addr, length, cfg));
			break;
		case hyperlink:
			return (mpm_transport_hyplnk_mmap(sp, addr, length, cfg));
			break;
		default:
			mpm_printf(1, "unsupported transport %d\n", sp->transport);
	}
	return (void *) -1;
}

/*
	Creates a mmap for the memory location and returns virtual address.
*/
uint32_t mpm_transport_phys_map(mpm_transport_h h, uint32_t addr, uint32_t length, mpm_transport_mmap_t *cfg)
{
	mpm_transport_cfg_t *sp = (mpm_transport_cfg_t *)h;
	if (!h) {
		mpm_printf(1, "invalid handle \n");
		return -1;
	}
	switch (sp->transport) {
		case shared_memory:
			mpm_printf(1, "shared_mem does not support phys map\n");
			break;
		case hyperlink:
			return (mpm_transport_hyplnk_phys_map(sp, addr, length, cfg));
			break;
		default:
			mpm_printf(1, "unsupported transport %d\n", sp->transport);
	}
	return -1;
}

/*
	Creates a mmap for the memory location and returns virtual address.
*/
void mpm_transport_munmap(mpm_transport_h h, void *va, uint32_t length)
{
	mpm_transport_cfg_t *sp = (mpm_transport_cfg_t *)h;
	if (!h) {
		mpm_printf(1, "invalid handle \n");
		return;
	}
	switch (sp->transport) {
		case shared_memory:
			if (mpm_transport_sharedmem_munmap(sp, va, length)) {
				mpm_printf(1, "shm munmap failed\n");
				return;
			}
			break;
		case hyperlink:
			if (mpm_transport_hyplnk_munmap(sp, va, length)) {
				mpm_printf(1, "shm munmap failed\n");
				return;
			}
			break;
		default:
			mpm_printf(1, "unsupported transport %d\n", sp->transport);
			return;
	}
}

/*
	Close mct connection.
*/
void mpm_transport_close(mpm_transport_h h)
{
	mpm_transport_cfg_t *sp = (mpm_transport_cfg_t *)h;
	if (!h) {
		mpm_printf(1, "invalid handle \n");
		return;
	}
	switch (sp->transport) {
		case shared_memory:
			mpm_transport_sharedmem_close(sp);
			break;
		case hyperlink:
			mpm_transport_hyplnk_close(sp);
			break;
		default:
			mpm_printf(1, "unsupported transport %d\n", sp->transport);
			return;
	}

	free(sp);
}

/*
	Creates a mmap for an extended 36-bit memory location and returns 32-bit virtual address.
*/
uint32_t mpm_transport_phys_map64(mpm_transport_h h, uint64_t addr, uint32_t length, mpm_transport_mmap_t *cfg)
{
	mpm_transport_cfg_t *sp = (mpm_transport_cfg_t *)h;
	if (!h) {
		mpm_printf(1, "invalid handle \n");
		return -1;
	}
	switch (sp->transport) {
		case shared_memory:
			mpm_printf(1, "shared_mem does not support phys map\n");
			break;
		case hyperlink:
			return (mpm_transport_hyplnk_phys_map64(sp, addr, length, cfg));
			break;
		default:
			mpm_printf(1, "unsupported transport %d\n", sp->transport);
	}
	return -1;
}

/*
	Similar to mpm_transport_read, except memcpy is replaced with EDMA3 operation.
	Returns 0 on success, error code on error.
*/
mpm_transport_trans_h mpm_transport_get_initiate64(mpm_transport_h h, uint64_t from_addr, uint64_t to_addr, uint32_t length, bool is_blocking, mpm_transport_read_t *cfg)
{
	mpm_transport_cfg_t *sp = (mpm_transport_cfg_t *)h;
	mpm_transport_trans_t *tp = (mpm_transport_trans_t *) NULL;
	if (!h) {
		mpm_printf(1, "invalid handle \n");
		return tp;
	}
	switch (sp->transport) {
		case shared_memory:
			mpm_printf(1, "shared_mem dma functions not supported\n");
			break;
		case hyperlink:
			tp = mpm_transport_hyplnk_get_initiate64(sp, from_addr, to_addr, length, is_blocking, cfg);
			if (tp->key == -1)
				mpm_printf(1, "hyplnk dma read failed\n");
			else
				mpm_printf(2, "got key: 0x%08x\n",tp->key);
			break;
		default:
			mpm_printf(1, "unsupported transport %d\n", sp->transport);
	}
	return tp;
}

/*
	Similar to mpm_transport_write, except memcpy is replaced with EDMA3 operation.
	Returns 0 on success, error code on error.
*/
mpm_transport_trans_h mpm_transport_put_initiate64(mpm_transport_h h, uint64_t to_addr, uint64_t from_addr, uint32_t length, bool is_blocking, mpm_transport_write_t *cfg)
{
	mpm_transport_cfg_t *sp = (mpm_transport_cfg_t *)h;
	mpm_transport_trans_t *tp = (mpm_transport_trans_t *) NULL;
	if (!h) {
		mpm_printf(1, "invalid handle \n");
		return tp;
	}
	switch (sp->transport) {
		case shared_memory:
			mpm_printf(1, "shared_mem dma functions not supported\n");
			break;
		case hyperlink:
			tp = mpm_transport_hyplnk_put_initiate64(sp, to_addr, from_addr, length, is_blocking, cfg);
			if (tp->key == -1)
				mpm_printf(1, "hyplnk dma write failed\n");
			else
				mpm_printf(2, "got key: 0x%08x\n",tp->key);
			break;
		default:
			mpm_printf(1, "unsupported transport %d\n", sp->transport);
	}
	return tp;
}

void mpm_printf(int val, const char* str, ...) {
	if (VERBOSITY_LEVEL == 0) return;
	if (val <= VERBOSITY_LEVEL) {
		va_list args;
		va_start(args, str);
		vprintf(str, args);
		va_end(args);
	}
}

int mpm_transport_peripheral_enable(char *slave_name, mpm_transport_open_t *cfg)
{
	mpm_transport_cfg_t *sp = NULL;
	char *config_buf = NULL;
	cJSON *json = NULL;
	int retval = 0;
	FILE *fpr = NULL;
	struct stat buf;
	int flen;
	int res;
	
	if (!slave_name || !cfg) {
		mpm_printf(1, "Null parameter\n");
		goto close_n_exit;
	}
	if (strlen(slave_name) > MPM_MAX_NAME_LENGTH) {
		mpm_printf(1, "Name should be less than %d\n", MPM_MAX_NAME_LENGTH);
		goto close_n_exit;
	}

	sp = calloc(1, sizeof(mpm_transport_cfg_t));
	if (!sp) {
		mpm_printf(1, "can't allocate memory for handle\n");
		goto close_n_exit;
	}

	fpr = fopen(DEFAULT_CONFIG_FILE_NAME, "r");
	if (!fpr){
		mpm_printf(1, "Error opening file %s (error: %s)\n",
				DEFAULT_CONFIG_FILE_NAME, strerror(errno));
		goto close_n_exit;
	}

	fstat(fileno(fpr), &buf);
	flen = buf.st_size;

	config_buf = (char *) malloc(flen + 1);
	if (!config_buf) {
		mpm_printf(1, "Error in allocating cfg buffer (error: %s)\n",
				strerror(errno));
		goto close_n_exit;
	}
	config_buf[flen] = 0;

	res = fread (config_buf, 1, flen, fpr);
	if (res != flen) {
		mpm_printf(1, "Error in reading cfg file (error: %s)\n",
				strerror(errno));
		goto close_n_exit;
	}

	json = cJSON_Parse(config_buf);
	if (!json) {
		mpm_printf(1, "JSON parser error before: [%s]\n",cJSON_GetErrorPtr());
		goto close_n_exit;
	}

	if (json_get_slave_info(json, slave_name, sp)) {
		mpm_printf(1, "Error in JSON parsing\n");
		goto close_n_exit;
	}

	switch (sp->transport) {
		case shared_memory:
			mpm_printf(2, "No special routine needed to enable shared_memory\n");
			goto close_n_exit;
			break;
		case hyperlink:
			if (retval = mpm_transport_hyplnk_enable_peripheral(sp, cfg)) {
				mpm_printf(1, "hyplnk peripheral enable failed\n");
				goto close_n_exit;
			}
			break;
		default:
			mpm_printf(1, "unsupported transport %d\n", sp->transport);
			goto close_n_exit;
	}


close_n_exit:
	if(json) cJSON_Delete(json);
	if (config_buf) free(config_buf);
	if (fpr) fclose(fpr);
	free(sp);

	return retval;
}

int mpm_transport_peripheral_disable(char *slave_name)
{
	mpm_transport_cfg_t *sp = NULL;
	char *config_buf = NULL;
	cJSON *json = NULL;
	int retval = 0;
	FILE *fpr = NULL;
	struct stat buf;
	int flen;
	int res;
	
	if (!slave_name) {
		mpm_printf(1, "Null parameter\n");
		goto close_n_exit;
	}
	if (strlen(slave_name) > MPM_MAX_NAME_LENGTH) {
		mpm_printf(1, "Name should be less than %d\n", MPM_MAX_NAME_LENGTH);
		goto close_n_exit;
	}

	sp = calloc(1, sizeof(mpm_transport_cfg_t));
	if (!sp) {
		mpm_printf(1, "can't allocate memory for handle\n");
		goto close_n_exit;
	}

	fpr = fopen(DEFAULT_CONFIG_FILE_NAME, "r");
	if (!fpr){
		mpm_printf(1, "Error opening file %s (error: %s)\n",
				DEFAULT_CONFIG_FILE_NAME, strerror(errno));
		goto close_n_exit;
	}

	fstat(fileno(fpr), &buf);
	flen = buf.st_size;

	config_buf = (char *) malloc(flen + 1);
	if (!config_buf) {
		mpm_printf(1, "Error in allocating cfg buffer (error: %s)\n",
				strerror(errno));
		goto close_n_exit;
	}
	config_buf[flen] = 0;

	res = fread (config_buf, 1, flen, fpr);
	if (res != flen) {
		mpm_printf(1, "Error in reading cfg file (error: %s)\n",
				strerror(errno));
		goto close_n_exit;
	}

	json = cJSON_Parse(config_buf);
	if (!json) {
		mpm_printf(1, "JSON parser error before: [%s]\n",cJSON_GetErrorPtr());
		goto close_n_exit;
	}

	if (json_get_slave_info(json, slave_name, sp)) {
		mpm_printf(1, "Error in JSON parsing\n");
		goto close_n_exit;
	}

	switch (sp->transport) {
		case shared_memory:
			mpm_printf(2, "No special routine needed to disable shared_memory\n");
			goto close_n_exit;
			break;
		case hyperlink:
			if (retval = mpm_transport_hyplnk_disable_peripheral(sp)) {
				mpm_printf(1, "hyplnk peripheral disable failed\n");
				goto close_n_exit;
			}
			break;
		default:
			mpm_printf(1, "unsupported transport %d\n", sp->transport);
			goto close_n_exit;
	}


close_n_exit:
	if(json) cJSON_Delete(json);
	if (config_buf) free(config_buf);
	if (fpr) fclose(fpr);
	free(sp);

	return retval;
}
